package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.query.JsonMath;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.util.Util;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public record JQAsExpression(
        @NotNull JQExpression variable, @NotNull List<@NotNull Pattern> patterns, @NotNull JQExpression expression
) implements JQExpression {

    public JQAsExpression {
        Objects.requireNonNull(variable);
        Objects.requireNonNull(expression);

        patterns = List.copyOf(patterns);
    }

    @Override
    public @NotNull Stream<@Nullable JsonValue> evaluate(@NotNull Context context) {
        var variables = new HashMap<String, JsonValue>();
        patterns.stream().map(Pattern::variables).flatMap(Set::stream).forEach(key -> variables.put(key, null));

        return variable.evaluate(context)
                .flatMap(value -> {
                    Stream<Map<String, JsonValue>> result = null;

                    // find first pattern that does not throw
                    var it = patterns.iterator();
                    while (it.hasNext()) {
                        try {
                             result = it.next().bind(context, value);
                        } catch (JsonQueryException ex) {
                            if (!it.hasNext()) throw ex;
                        }
                    }

                    // execute expression for all possible pattern matches
                    assert result != null;
                    return result
                            .map(vars -> {
                                var out = new HashMap<>(variables);
                                out.putAll(vars);
                                return context.withVariables(out);
                            })
                            .flatMap(expression::evaluate);
                });
    }

    @Override
    public boolean isConstant() {
        return variable.isConstant() && expression.isConstant();
    }

    @Override
    public @NotNull String toString() {
        return variable
               + " as " + patterns.stream().map(Objects::toString).collect(Collectors.joining(" ?// "))
               + " | " + expression;
    }

    public sealed interface Pattern {
        @NotNull Set<@NotNull String> variables();
        @NotNull Stream<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value);

        record ValuePattern(@NotNull String name) implements Pattern {
            public ValuePattern {
                Objects.requireNonNull(name);
                if (!name.startsWith("$")) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                return Set.of(name);
            }

            @Override
            public @NotNull Stream<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var map = new HashMap<String, JsonValue>();
                map.put(name, value);
                return Stream.of(Collections.unmodifiableMap(map));
            }

            @Override
            public @NotNull String toString() {
                return name;
            }
        }

        record ArrayPattern(@NotNull List<@NotNull Pattern> patterns) implements Pattern {
            public ArrayPattern {
                patterns = List.copyOf(patterns);
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Stream<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Stream<Map<String, JsonValue>>>>();

                for (int i = 0; i < patterns.size(); i++) {
                    var k = new JsonNumber(i);
                    var v = JsonMath.index(value, k);

                    var pattern = patterns.get(i);
                    streams.add(() -> pattern.bind(context, v));
                }

                return Util.crossReversed(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                return patterns.stream().map(Objects::toString).collect(Collectors.joining(", ", "[", "]"));
            }
        }

        record ObjectPattern(@NotNull SequencedMap<@NotNull JQExpression, @NotNull Pattern> patterns) implements Pattern {

            public ObjectPattern {
                patterns = Collections.unmodifiableSequencedMap(new LinkedHashMap<>(patterns));
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.values().forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Stream<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Stream<Map<String, JsonValue>>>>();

                this.patterns.reversed().forEach((keyExpression, pattern) -> streams.add(() -> {
                    var keyStream = keyExpression.evaluate(context);
                    return keyStream.flatMap(key -> pattern.bind(context, JsonMath.index(value, key)));
                }));

                return Util.cross(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.reversed().forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                var out = new StringJoiner(", ", "{", "}");
                patterns.forEach((key, pattern) -> out.add(key + ": " + pattern));
                return out.toString();
            }
        }
    }
}
