package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.query.JsonMath;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.util.Util;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Gatherer;
import java.util.stream.Stream;

public enum JQBuiltIn implements JQInvocable {
    ABS(0, (context, _) -> context.stream().map(JsonMath::abs)),
    ADD(0, (context, _) -> context.stream().map(value -> JsonMath.values(value).reduce(null, JsonMath::add))),
    NOT(0, (context, _) -> context.stream().map(JsonMath::not)),

    // error handling
    ERROR$0(0, (context, _) -> context.stream().flatMap(JsonMath::error)),
    ERROR$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(JsonMath::error)),
    HALT(0, (_, _) -> Stream.of((JsonValue) null).flatMap(_ -> JsonMath.halt())),
    HALT_ERROR$0(0, (context, _) -> context.stream().flatMap(JsonMath::halt)),
    HALT_ERROR$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(value -> JsonMath.halt(context.root(), value))),

    // stream operations
    EMPTY(0, (_, _) -> Stream.empty()),
    RANGE$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(JsonMath::range)),
    RANGE$2(2, (context, args) -> Util.cross(args, context).flatMap(bounds -> JsonMath.range(bounds.get(0), bounds.get(1)))),
    RANGE$3(3, (context, args) -> Util.cross(args, context).flatMap(bounds -> JsonMath.range(bounds.get(0), bounds.get(1), bounds.get(2)))),
    LIMIT(2, (context, args) -> args.getFirst().evaluate(context).flatMap(limit -> JsonMath.limit(args.getLast().evaluate(context), limit))),
    FIRST$1(1, (context, args) -> JsonMath.first(args.getFirst().evaluate(context))),
    LAST$1(1, (context, args) -> JsonMath.last(args.getFirst().evaluate(context))),
    NTH$2(2, (context, args) -> args.getFirst().evaluate(context).flatMap(n -> JsonMath.nth(args.getLast().evaluate(context), n))),

    // iterable operations
    MAP(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.map(value, filter));
    }),
    MAP_VALUES(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.mapValues(value, filter));
    }),
    KEYS(0, (context, _) -> context.stream().map(JsonMath::keys)),
    KEYS_UNSORTED(0, (context, _) -> context.stream().map(JsonMath::keysUnsorted)),
    HAS(1, (context, args) -> args.getFirst().evaluate(context).map(index -> JsonMath.has(context.root(), index))),
    IN(1, (context, args) -> args.getFirst().evaluate(context).map(value -> JsonMath.in(context.root(), value))),
    FIRST$0(0, (context, _) -> context.stream().map(value -> JsonMath.index(value, JsonNumber.ZERO))),
    LAST$0(0, (context, _) -> context.stream().map(value -> JsonMath.index(value, new JsonNumber(-1)))),
    NTH$1(1, (context, args) -> args.getFirst().evaluate(context).map(index -> JsonMath.index(context.root(), index))),
    ANY$0(0, (context, _) -> context.stream().map(JsonMath::any)),
    ANY$1(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.any(value, filter));
    }),
    ANY$2(2, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return Util.lazy(() -> JsonMath.any(args.getFirst().evaluate(context), filter));
    }),
    ALL$0(0, (context, _) -> context.stream().map(JsonMath::all)),
    ALL$1(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.all(value, filter));
    }),
    ALL$2(2, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return Util.lazy(() -> JsonMath.all(args.getFirst().evaluate(context), filter));
    }),
    FLATTEN$0(0, (context, _) -> context.stream().map(JsonMath::flatten)),
    FLATTEN$1(1, (context, args) -> context.stream().flatMap(value -> args.getFirst().evaluate(context).map(depth -> JsonMath.flatten(value, depth)))),
    SORT(0, (context, _) -> context.stream().map(JsonMath::sort)),
    SORT_BY(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.sort(value, filter));
    }),
    MIN(0, (context, _) -> context.stream().map(JsonMath::min)),
    MIN_BY(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.min(value, filter));
    }),
    MAX(0, (context, _) -> context.stream().map(JsonMath::max)),
    MAX_BY(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.max(value, filter));
    }),
    UNIQUE(0, (context, _) -> context.stream().map(JsonMath::unique)),
    UNIQUE_BY(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.unique(value, filter));
    }),
    GROUP_BY(1, (context, args) -> {
        var filter = args.getFirst().bind(context);
        return context.stream().map(value -> JsonMath.group(value, filter));
    }),
    REVERSE(0, (context, _) -> context.stream().map(JsonMath::reverse)),
    CONTAINS(1, (context, args) -> args.getFirst().evaluate(context).map(content -> JsonMath.contains(context.root(), content))),
    INDICES(1, (context, args) -> args.getFirst().evaluate(context).map(content -> JsonMath.indices(context.root(), content))),
    INDEX(1, (context, args) -> args.getFirst().evaluate(context).map(content -> JsonMath.firstindex(context.root(), content))),
    RINDEX(1, (context, args) -> args.getFirst().evaluate(context).map(content -> JsonMath.lastindex(context.root(), content))),
    INSIDE(1, (context, args) -> args.getFirst().evaluate(context).map(content -> JsonMath.inside(context.root(), content))),
    COMBINATIONS$0(0, (context, _) -> context.stream().flatMap(JsonMath::combinations)),
    COMBINATIONS$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(n -> JsonMath.combinations(context.root(), n))),
    BSEARCH(1, (context, args) -> args.getFirst().evaluate(context).map(value -> JsonMath.bsearch(context.root(), value))),
    TRANSPOSE(0, (context, _) -> context.stream().map(JsonMath::transpose)),

    // filters
    ARRAYS(0, (context, _) -> context.stream().filter(JsonMath::isArray)),
    OBJECTS(0, (context, _) -> context.stream().filter(JsonMath::isObject)),
    ITERABLES(0, (context, _) -> context.stream().filter(JsonMath::isIterable)),
    BOOLEANS(0, (context, _) -> context.stream().filter(JsonMath::isBoolean)),
    NUMBERS(0, (context, _) -> context.stream().filter(JsonMath::isNumber)),
    NORMALS(0, (context, _) -> context.stream().filter(JsonMath::isNormal)),
    FINITES(0, (context, _) -> context.stream().filter(JsonMath::isFinite)),
    STRINGS(0, (context, _) -> context.stream().filter(JsonMath::isString)),
    NULLS(0, (context, _) -> context.stream().filter(JsonMath::isNull)),
    VALUES(0, (context, _) -> context.stream().filter(JsonMath::isValue)),
    SCALARS(0, (context, _) -> context.stream().filter(JsonMath::isScalar)),

    // checks
    ISINFINITE(0, (context, _) -> context.stream().map(JsonMath::isInfinite).map(JsonBoolean::valueOf)),
    ISNAN(0, (context, _) -> context.stream().map(JsonMath::isNan).map(JsonBoolean::valueOf)),
    ISFINITE(0, (context, _) -> context.stream().map(JsonMath::isFinite).map(JsonBoolean::valueOf)),
    ISNORMAL(0, (context, _) -> context.stream().map(JsonMath::isNormal).map(JsonBoolean::valueOf)),
    ISEMPTY(1, (context, args) -> Util.lazy(() -> JsonBoolean.valueOf(JsonMath.isEmpty(args.getFirst().evaluate(context))))),

    // string operations
    TRIM(0, (context, _) -> context.stream().map(JsonMath::trim)),
    LTRIM(0, (context, _) -> context.stream().map(JsonMath::ltrim)),
    RTRIM(0, (context, _) -> context.stream().map(JsonMath::rtrim)),
    LTRIMSTR(1, (context, args) -> args.getFirst().evaluate(context).map(prefix -> JsonMath.ltrimstr(context.root(), prefix))),
    RTRIMSTR(1, (context, args) -> args.getFirst().evaluate(context).map(suffix -> JsonMath.rtrimstr(context.root(), suffix))),
    SPLIT$1(1, (context, args) -> args.getFirst().evaluate(context).map(separator -> JsonMath.split(context.root(), separator))),
    JOIN(1, (context, args) -> args.getFirst().evaluate(context).map(separator -> JsonMath.join(context.root(), separator))),
    IMPLODE(0, (context, _) -> context.stream().map(JsonMath::implode)),
    EXPLODE(0, (context, _) -> context.stream().map(JsonMath::explode)),
    ASCII_UPCASE(0, (context, _) -> context.stream().map(JsonMath::asciiUpcase)),
    ASCII_DOWNCASE(0, (context, _) -> context.stream().map(JsonMath::asciiDowncase)),
    UTF8BYTELENGTH(0, (context, _) -> context.stream().map(JsonMath::utf8ByteLength)),
    STARTSWITH(1, (context, args) -> args.getFirst().evaluate(context).map(prefix -> JsonMath.startswith(context.root(), prefix))),
    ENDSWITH(1, (context, args) -> args.getFirst().evaluate(context).map(prefix -> JsonMath.endswith(context.root(), prefix))),

    // regex
    TEST$1(1, JsonMath::test),
    TEST$2(2, JsonMath::test),
    MATCH$1(1, JsonMath::match),
    MATCH$2(2, JsonMath::match),
    CAPTURE$1(1, JsonMath::capture),
    CAPTURE$2(2, JsonMath::capture),
    SCAN$1(1, JsonMath::scan),
    SCAN$2(2, JsonMath::scan),
    SPLIT$2(2, JsonMath::split),
    SPLITS$1(1, JsonMath::splits),
    SPLITS$2(2, JsonMath::splits),
    SUB$2(2, JsonMath::sub),
    SUB$3(3, JsonMath::sub),
    GSUB$2(2, JsonMath::gsub),
    GSUB$3(3, JsonMath::gsub),

    // conversions
    TYPE(0, (context, _) -> context.stream().map(JsonMath::type)),
    TOJSON(0, (context, _) -> context.stream().map(JsonMath::tojson)),
    TOSTRING(0, (context, _) -> context.stream().map(JsonMath::tostring)),
    TONUMBER(0, (context, _) -> context.stream().map(JsonMath::tonumber)),
    FROMJSON(0, (context, _) -> context.stream().map(JsonMath::fromjson)),

    // misc
    LENGTH(0, (context, _) -> context.stream().map(JsonMath::length)),
    REPEAT(1, (context, args) -> Stream.generate(() -> args.getFirst().evaluate(context)).flatMap(Function.identity())),

    // math library
    ACOS(0, (context, _) -> context.stream().map(JsonMath::acos)),
    ACOSH(0, (context, _) -> context.stream().map(JsonMath::acosh)),
    ASIN(0, (context, _) -> context.stream().map(JsonMath::asin)),
    ASINH(0, (context, _) -> context.stream().map(JsonMath::asinh)),
    ATAN(0, (context, _) -> context.stream().map(JsonMath::atan)),
    ATANH(0, (context, _) -> context.stream().map(JsonMath::atanh)),
    CBRT(0, (context, _) -> context.stream().map(JsonMath::cbrt)),
    CEIL(0, (context, _) -> context.stream().map(JsonMath::ceil)),
    COS(0, (context, _) -> context.stream().map(JsonMath::cos)),
    COSH(0, (context, _) -> context.stream().map(JsonMath::cosh)),
    ERF(0, (context, _) -> context.stream().map(JsonMath::erf)),
    ERFC(0, (context, _) -> context.stream().map(JsonMath::erfc)),
    EXP(0, (context, _) -> context.stream().map(JsonMath::exp)),
    EXP10(0, (context, _) -> context.stream().map(JsonMath::exp10)),
    EXP2(0, (context, _) -> context.stream().map(JsonMath::exp2)),
    EXPM1(0, (context, _) -> context.stream().map(JsonMath::expm1)),
    FABS(0, (context, _) -> context.stream().map(JsonMath::fabs)),
    FLOOR(0, (context, _) -> context.stream().map(JsonMath::floor)),
    GAMMA(0, (context, _) -> context.stream().map(JsonMath::gamma)),
    J0(0, (context, _) -> context.stream().map(JsonMath::j0)),
    J1(0, (context, _) -> context.stream().map(JsonMath::j1)),
    LGAMMA(0, (context, _) -> context.stream().map(JsonMath::lgamma)),
    LOG(0, (context, _) -> context.stream().map(JsonMath::log)),
    LOG10(0, (context, _) -> context.stream().map(JsonMath::log10)),
    LOG1P(0, (context, _) -> context.stream().map(JsonMath::log1p)),
    LOG2(0, (context, _) -> context.stream().map(JsonMath::log2)),
    LOGB(0, (context, _) -> context.stream().map(JsonMath::logb)),
    NEARBYINT(0, (context, _) -> context.stream().map(JsonMath::nearbyint)),
    RINT(0, (context, _) -> context.stream().map(JsonMath::rint)),
    ROUND(0, (context, _) -> context.stream().map(JsonMath::round)),
    SIGNIFICAND(0, (context, _) -> context.stream().map(JsonMath::significand)),
    SIN(0, (context, _) -> context.stream().map(JsonMath::sin)),
    SINH(0, (context, _) -> context.stream().map(JsonMath::sinh)),
    SQRT(0, (context, _) -> context.stream().map(JsonMath::sqrt)),
    TAN(0, (context, _) -> context.stream().map(JsonMath::tan)),
    TANH(0, (context, _) -> context.stream().map(JsonMath::tanh)),
    TGAMMA(0, (context, _) -> context.stream().map(JsonMath::tgamma)),
    TRUNC(0, (context, _) -> context.stream().map(JsonMath::trunc)),
    Y0(0, (context, _) -> context.stream().map(JsonMath::y0)),
    Y1(0, (context, _) -> context.stream().map(JsonMath::y1)),
    ATAN2(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.atan2(values.get(0), values.get(1)))),
    COPYSIGN(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.copysign(values.get(0), values.get(1)))),
    DREM(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.drem(values.get(0), values.get(1)))),
    FDIM(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fdim(values.get(0), values.get(1)))),
    FMAX(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fmax(values.get(0), values.get(1)))),
    FMIN(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fmin(values.get(0), values.get(1)))),
    FMOD(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fmod(values.get(0), values.get(1)))),
    FREXP(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.frexp(values.get(0), values.get(1)))),
    HYPOT(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.hypot(values.get(0), values.get(1)))),
    JN(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.jn(values.get(0), values.get(1)))),
    LDEXP(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.ldexp(values.get(0), values.get(1)))),
    MODF(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.modf(values.get(0), values.get(1)))),
    NEXTAFTER(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.nextafter(values.get(0), values.get(1)))),
    NEXTTOWARD(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.nexttoward(values.get(0), values.get(1)))),
    POW(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.pow(values.get(0), values.get(1)))),
    REMAINDER(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.remainder(values.get(0), values.get(1)))),
    SCALB(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.scalb(values.get(0), values.get(1)))),
    SCALBLN(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.scalbln(values.get(0), values.get(1)))),
    YN(2, (context, args) -> Util.cross(args, context).map(values -> JsonMath.yn(values.get(0), values.get(1)))),
    FMA(3, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fma(values.get(0), values.get(1), values.get(2))))
    ;

    public static final @NotNull Map<@NotNull String, @NotNull JQInvocable> ALL_BUILTINS = Arrays.stream(JQBuiltIn.values())
            .collect(Collectors.toMap(
                    JQInvocable::reference,
                    Function.identity()
            ));

    private final @NotNull String identifier;
    private final int arity;
    private final @NotNull Implementation implementation;

    JQBuiltIn(int arity, @NotNull Implementation implementation) {
        var identifier = this.name().toLowerCase(Locale.ROOT);
        var idx = identifier.lastIndexOf("$");
        if (idx != -1) identifier = identifier.substring(0, idx);

        this.identifier = identifier;
        this.arity = arity;
        this.implementation = implementation;
    }

    @Override
    public @NotNull Stream<@Nullable JsonValue> invoke(JQExpression.@NotNull Context context, @NotNull List<@NotNull JQExpression> args) {
        if (args.size() != arity) throw new JsonQueryException("invalid argument count");
        return implementation.invoke(context, args);
    }

    @Override
    public @NotNull String identifier() {
        return identifier;
    }

    @Override
    public int arity() {
        return arity;
    }

    @FunctionalInterface
    private interface Implementation {
        @NotNull Stream<@Nullable JsonValue> invoke(@NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args);
    }
}
