package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonString;
import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.JsonQueryHaltException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public record JQTryExpression(@NotNull JQExpression expression, @Nullable JQExpression fallback) implements JQExpression {
    public JQTryExpression {
        Objects.requireNonNull(expression);
    }

    public JQTryExpression(@NotNull JQExpression expression) {
        this(expression, null);
    }

    @Override
    public @NotNull Stream<@Nullable JsonValue> evaluate(@NotNull Context context) {
        var iterator = new QuietIterator(
                expression.evaluate(context).iterator(),
                fallback == null ? null : ex -> fallback.evaluate(context.withRoot(new JsonString(ex.getMessage()))).iterator()
        );
        var spliterator = Spliterators.spliteratorUnknownSize(iterator, Spliterator.ORDERED);
        return StreamSupport.stream(spliterator, false);
    }

    @Override
    public boolean isConstant() {
        return expression.isConstant() && (fallback == null || fallback.isConstant());
    }

    @Override
    public @NotNull String toString() {
        return "try " + expression + (fallback == null ? "" : " catch " + fallback);
    }

    private static class QuietIterator implements Iterator<@Nullable JsonValue> {
        private @Nullable Iterator<@Nullable JsonValue> delegate;
        private @Nullable Function<@NotNull JsonQueryException, @NotNull Iterator<@Nullable JsonValue>> fallback;

        private @Nullable JsonValue next;
        private boolean hasNext;

        private boolean valid = false;

        private QuietIterator(
                @NotNull Iterator<@Nullable JsonValue> delegate,
                @Nullable Function<@NotNull JsonQueryException, @NotNull Iterator<@Nullable JsonValue>> fallback
        ) {
            this.delegate = delegate;
            this.fallback = fallback;
        }

        private void ensureValid() {
            if (valid) return;

            while (true) {
                try {
                    if (delegate != null && delegate.hasNext()) { // still have values in current stream
                        next = delegate.next();
                        hasNext = true;
                    } else { // end of stream
                        next = null;
                        hasNext = false;
                    }
                    break;
                } catch (JsonQueryHaltException ex) {
                    throw ex;
                } catch (JsonQueryException ex) { // switch to fallback
                    delegate = fallback != null ? fallback.apply(ex) : null;
                    fallback = null;
                }
            }
            valid = true;
        }

        @Override
        public boolean hasNext() {
            ensureValid();
            return hasNext;
        }

        @Override
        public @Nullable JsonValue next() {
            ensureValid();
            if (!hasNext) throw new NoSuchElementException();
            valid = false;
            return next;
        }
    }
}
